/*-
 * Copyright (c) 1998 John Galbraith
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* This is a driver for the Mark of the Unicorn Midi Express XT.  The
   driver version is 1.0, written Nov 8-10, 1998. */

#include "mtp.h"

#if NMTP > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/fcntl.h>
#include <sys/uio.h>
#include </sys/i386/include/clock.h>

#include </sys/dev/ppbus/ppbconf.h>

struct mtp_data {

    int		mtp_unit;
    int		mtp_flags;
#define HAVE_PPBUS	(1<<0)
    struct ppb_device mtp_dev;
    short asleep;
    short bytes_left;
    u_int8_t running_status;
    u_int8_t running_cable;
    u_int8_t lastbyte;
    short msg_length;
    short open;

        /* hold all data read in circular buffer */
#define BUFSIZE 2048
    u_int8_t buf[BUFSIZE];
    int bufend;

        /* maintain another circular buffer of message pointers */
#define MSGBUFSIZE 32
    int msgptr[MSGBUFSIZE];
    int firstmessage; /* pointer to oldest complete message */
    int lastmessage;  /* pointer to next message under construction */
};

#define queue_byte(mtp, byte) \
     (mtp)->buf[(mtp)->bufend++] = (byte); \
     if((mtp)->bufend == BUFSIZE) \
         (mtp)->bufend = 0; \
     (mtp)->lastbyte = (byte)

#define complete_message(mtp) \
     (mtp)->lastmessage += 1; \
     if((mtp)->lastmessage == MSGBUFSIZE) \
         (mtp)->lastmessage = 0; \
     (mtp)->msgptr[(mtp)->lastmessage] = (mtp)->bufend;

    /* mark a message as read */
#define pop_message(mtp) {\
     (mtp)->firstmessage++; \
     if((mtp)->firstmessage == MSGBUFSIZE) \
         (mtp)->firstmessage = 0; \
     }

#define MAXMTP		2			/* XXX not much better! */
#define	MTP_PRI		(PZERO+8)|PCATCH
static int 		nmtp = 0;
static struct mtp_data	*mtpdata[MAXMTP];
#define MTP_MODULE

#define MTP_WRITE      0x04
#define MTP_READ       0x08
#define MTP_IRQ_ENABLE 0x10

#define MTP_BYTE_AVAIL 0x08
#define MTP_DATA0      0x10
#define MTP_DATA1      0x20
#define MTP_IRQ        0x40
#define MTP_RFD        0x80

/*
 * Make ourselves visible as a ppbus driver
 */

static struct ppb_device	*mtpprobe(struct ppb_data *ppb);
static int			mtpattach(struct ppb_device *dev);
static void			mtpintr(int unit);

static struct ppb_driver mtpdriver = {
    mtpprobe, mtpattach, "mtp"
};
DATA_SET(ppbdriver_set, mtpdriver);

static	d_open_t	mtpopen;
static	d_close_t	mtpclose;
static	d_ioctl_t	mtpioctl;
static  d_write_t       mtpwrite;
static  d_read_t        mtpread;

#define CDEV_MAJOR 20
static struct cdevsw mtp_cdevsw = 
{ mtpopen,	mtpclose,	mtpread,	mtpwrite,  /* 20 */
  mtpioctl,	nullstop,	nullreset,	nodevtotty,
  seltrue,	nommap,		nostrat,	"mtp",	NULL,	-1 };

/*
 * mtpprobe()
 */
static struct ppb_device *
mtpprobe(struct ppb_data *ppb)
{
    struct mtp_data *mtp;

    mtp = (struct mtp_data *) malloc(sizeof(struct mtp_data),
                                     M_TEMP, M_NOWAIT);
    if (!mtp) {
        printf("mtp: cannot malloc!\n");
        return 0;
    }
    bzero(mtp, sizeof(struct mtp_data));

    mtpdata[nmtp] = mtp;

	/*
	 * mtp dependent initialisation.
	 */
    mtp->mtp_unit = nmtp;
    mtp->open = 0;
    mtp->bufend = 0;
    mtp->lastmessage = 0;
    mtp->firstmessage = 0;
    mtp->msgptr[0] = 0;
    mtp->bytes_left = -1;

	/*
	 * ppbus dependent initialisation.
	 */
    mtp->mtp_dev.id_unit = mtp->mtp_unit;
    mtp->mtp_dev.ppb = ppb;
    mtp->mtp_dev.intr = mtpintr;

	/* Ok, go to next device on next probe */
    nmtp ++;

    return &mtp->mtp_dev;
}

static int
mtpattach(struct ppb_device *dev)
{
    struct mtp_data *mtp = mtpdata[dev->id_unit];
    printf("mtp%d: <Midi Express XT> on ppbus %d\n",
           dev->id_unit, dev->ppb->ppb_link->adapter_unit);
    return (1);
}

static void
mtpintr(int unit)
{
    u_int8_t status, control;
    struct mtp_data *mtp = mtpdata[unit];

        /* FIXME! how do I shut these off? Or, are they some sort of
           useful timecode? */
    status = ppb_rstr(&mtp->mtp_dev);
    if(!(status & MTP_BYTE_AVAIL))
        return;

    control = ppb_rctr(&mtp->mtp_dev);
    do {
        int i;
        u_int8_t byte = 0;
            /* assemble byte, 2 bits at a time */
        for(i = 0; i < 8; i += 2)
        {
                /* assert C3, tells mtp to give status */
            control |= MTP_READ;
            ppb_wctr(&mtp->mtp_dev, control);
                /* read the status currently asserted */
            status = ppb_rstr(&mtp->mtp_dev);
                /* unassert C3, get ready for next 2 bits */
            control &= ~MTP_READ;
            ppb_wctr(&mtp->mtp_dev, control);
                /* here is the prize */
            byte |= ((status&(MTP_DATA0|MTP_DATA1))>>4)<<i;
        }

        switch(byte&0xF0)
        {
            case 0x80:
            case 0x90:
            case 0xA0:
            case 0xB0:
            case 0xE0:
            case 0xC0:
            case 0xD0:
                mtp->msg_length = 2;
                if((byte&0xF0 == 0xC0) || (byte&0xF0) == 0xD0)
                    mtp->msg_length = 1;
                mtp->bytes_left = mtp->msg_length;
                mtp->running_status = byte;
                if(mtp->msgptr[mtp->lastmessage] == mtp->bufend)
                {
                        /* running cable status */
                    queue_byte(mtp, 0xF5);
                    queue_byte(mtp, mtp->running_cable);
                }
                queue_byte(mtp, byte);
                break;
            case 0xF0:
                mtp->bytes_left = -1;  /* unknown quantity */
                switch(byte&0x0F)
                {
                    case 0x0E:  /* active sensing */
                        break;
                    case 0x05:
                        queue_byte(mtp, byte);
                        break;
                    case 0x07:  /* end of SYSEX */
                        queue_byte(mtp, byte);
                        complete_message(mtp);
                        break;
                    default:
                        queue_byte(mtp, byte);
                }
                break;
            default: /* must be data byte */
                if(mtp->lastbyte == 0xF5)
                    mtp->running_cable = byte;
                if(mtp->bytes_left == 0) /* must be running status */
                {
                    queue_byte(mtp, 0xF5);
                    queue_byte(mtp, mtp->running_cable);
                    queue_byte(mtp, mtp->running_status);
                    mtp->bytes_left = mtp->msg_length;
                }
                queue_byte(mtp, byte);
                mtp->bytes_left--;
                if(mtp->bytes_left == 0)
                {
                    complete_message(mtp);
                }
        }
            /* on the last byte, S3 is cleared after C3 is
               unasserted.  S3 must be read again here to know if
               there is another byte available */
        status = ppb_rstr(&mtp->mtp_dev);
    } while(status & MTP_BYTE_AVAIL);

    if ((mtp->asleep) && (mtp->firstmessage != mtp->lastmessage))
    {
        mtp->asleep = 0;
        wakeup(mtp);
    }
}

static int
mtpopen(dev_t dev, int flags, int fmt, struct proc *p)
{
    u_int unit = minor(dev);
    struct mtp_data *mtp = mtpdata[unit];
    int res;
    
    if (unit >= nmtp)
        return ENXIO;

    if(mtp->open)
        return EBUSY;
    
    if (!(mtp->mtp_flags & HAVE_PPBUS))
        if ((res = ppb_request_bus(&mtp->mtp_dev, (flags & O_NONBLOCK) ? PPB_DONTWAIT : (PPB_WAIT | PPB_INTR))))
            return (res);
    
    mtp->mtp_flags |= HAVE_PPBUS;

        /* enable interrupts with C4 */
    ppb_wctr(&mtp->mtp_dev, MTP_WRITE | MTP_IRQ_ENABLE);
    mtp->open = 1;
    return (0);
}

static int
mtpclose(dev_t dev, int flags, int fmt, struct proc *p)
{
    u_int unit = minor(dev);
    struct mtp_data *mtp = mtpdata[unit];
    
        /* disable interrupts */
    ppb_wctr(&mtp->mtp_dev, 0);

    if (mtp->mtp_flags & HAVE_PPBUS)
        ppb_release_bus(&mtp->mtp_dev);
    mtp->mtp_flags &= ~HAVE_PPBUS;
    mtp->open = 0;
    return (0);
}

static int mtpread(dev_t dev, struct uio *uio, int ioflag)
{
    struct mtp_data* mtp = mtpdata[minor(dev)];
    int s, msgstart, msgend, msgsize;

    s = spltty();
    if(mtp->firstmessage == mtp->lastmessage)  /* no message available */
    {
        mtp->asleep = 1;
        switch(tsleep(mtp, MTP_PRI, "read", 0))
        {
            case 0: /* normal interrupt wakeup */
                break;
            case EWOULDBLOCK: /* kernel timeout */
            case EINTR: /* interrupt system call */
            case ERESTART:
            default:
                return EIO;
                break;
        }
    }

    msgstart = mtp->msgptr[mtp->firstmessage];
    if(mtp->firstmessage+1 < MSGBUFSIZE)
        msgend = mtp->msgptr[mtp->firstmessage+1];
    else
        msgend = mtp->msgptr[0];

    if(msgend > msgstart)
        uiomove(mtp->buf + msgstart, msgend-msgstart, uio);
    else
    {
        uiomove(mtp->buf + msgstart, BUFSIZE-msgstart, uio);
        uiomove(mtp->buf, msgend, uio);
    }
    pop_message(mtp);
    splx(s);

    return 0;
}

static int mtpwrite(dev_t dev, struct uio *uio, int ioflag)
{
    int count, size, time;
    int unit = minor(dev);
    unsigned char buf[1024];
    int maxcount = 1024;
    u_int8_t control, status;

    struct mtp_data* mtp = mtpdata[unit];

    size = uio->uio_resid;
    uiomove(buf, maxcount, uio);

    control = ppb_rctr(&mtp->mtp_dev);
        /* wait for RFD */
    time = 1000;
    status = ppb_rstr(&mtp->mtp_dev);
    while(!(status & MTP_RFD) && (time-- > 0))
        status = ppb_rstr(&mtp->mtp_dev);
    if(!time)
        return EIO;

    for(count = 0; count < size; count++)
    {
        int time;

        ppb_wdtr(&mtp->mtp_dev, buf[count]);
        
            /* assert strobe */
        control &= ~MTP_WRITE;
        ppb_wctr(&mtp->mtp_dev, control);
            /* unassert strobe */
        control |= MTP_WRITE;
        ppb_wctr(&mtp->mtp_dev, control);
    }
    
    return 0;
}

static int
mtpioctl(dev_t dev, u_long cmd, caddr_t data, int flags, struct proc *p)
{
    u_int unit = minor(dev);
    struct mtp_data *mtp = mtpdata[unit];
    int error = 0;
    u_int8_t *val = (u_int8_t *)data;

    switch (cmd) 
    {
        default:
            error = ENOTTY;
            break;
    }
    return (error);
}

static mtp_devsw_installed = 0;

static void mtp_drvinit(void *unused)
{
    dev_t dev;

    if (!mtp_devsw_installed ) 
    {
        dev = makedev(CDEV_MAJOR, 0);
        cdevsw_add(&dev, &mtp_cdevsw, NULL);
        mtp_devsw_installed = 1;
    }
}

SYSINIT(mtpdev,SI_SUB_DRIVERS,SI_ORDER_MIDDLE+CDEV_MAJOR,mtp_drvinit,NULL)

#ifdef MTP_MODULE

#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/lkm.h>

MOD_DEV(mtp, LM_DT_CHAR, CDEV_MAJOR, &mtp_cdevsw);

static int
mtp_load(struct lkm_table *lkmtp, int cmd)
{
    struct ppb_data *ppb;
    struct ppb_device *dev;

    for (ppb = ppb_next_bus(NULL); ppb; ppb = ppb_next_bus(ppb))
    {
        dev = mtpprobe(ppb);
        mtpattach(dev);
        ppb_attach_device(dev);
    }

    return (0);
}

static int
mtp_unload(struct lkm_table *lkmtp, int cmd)
{
    int i;
    for(i = nmtp-1; i > 0; i--)
        if(mtpdata[i]->open)
            return 1;

    for(i = nmtp-1; i > 0; i--)
    {
        ppb_remove_device(&mtpdata[i]->mtp_dev);
        free(mtpdata[i], M_TEMP);
    }
    return (0);
}

int
mtp_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
    DISPATCH(mtp, lkmtp, cmd, ver, mtp_load, mtp_unload, lkm_nullcmd);
}

#endif /* MTP_MODULE */

#endif /* NMTP */
